//
//  main.cpp
//  05-IndirectRendering
//
//  Created by Orlando Pereira on 29/04/15.
//  Copyright (c) 2015 Orlando Pereira. All rights reserved.
//

// in order to get function prototypes from glext.h, define GL_GLEXT_PROTOTYPES before including glext.h
#define GL_GLEXT_PROTOTYPES

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <cstdlib>
#include <iostream>
#include <sstream>
#include <iomanip>
#include "glext.h"
#include "glInfo.h"

#include "objLoader.h"

using std::stringstream;
using std::cout;
using std::endl;
using std::ends;

void display();
void reshapeCB(int w, int h);
void idleCB();
void keyboardCB(unsigned char key, int x, int y);
void mouseCB(int button, int stat, int x, int y);
void mouseMotionCB(int x, int y);
void exitCB();

void drawString(const char *str, int x, int y, float color[4], void *font);
void showInfo();

void *font = GLUT_BITMAP_8_BY_13;
GLuint vboId = 0;
int screenWidth = 1024;
int screenHeight = 1024;
bool mouseLeftDown;
bool mouseRightDown;
float mouseX, mouseY, cameraAngleX, cameraAngleY;
float cameraDistance = 5.0f;
bool vboSupported, vboUsed;
int drawMode = 0;

// Cube
//    v6----- v5
//   /|      /|
//  v1------v0|
//  | |     | |
//  | |v7---|-|v4
//  |/      |/
//  v2------v3
//
// A cube has 6 sides and each side has 2 triangles.
// The cube consists of 36 vertices (6 sides * 2 tris * 3 vertices = 36 vertices).
// The size of vertex array is 108 floats (36 * 3 = 108).
GLfloat vertices[]  = { 1, 1, 1,  -1, 1, 1,  -1,-1, 1,  // v0-v1-v2 (front)
    -1,-1, 1,   1,-1, 1,   1, 1, 1, // v2-v3-v0
    1, 1, 1,   1,-1, 1,   1,-1,-1,  // v0-v3-v4 (right)
    1,-1,-1,   1, 1,-1,   1, 1, 1,  // v4-v5-v0
    1, 1, 1,   1, 1,-1,  -1, 1,-1,  // v0-v5-v6 (top)
    -1, 1,-1,  -1, 1, 1,   1, 1, 1, // v6-v1-v0
    -1, 1, 1,  -1, 1,-1,  -1,-1,-1, // v1-v6-v7 (left)
    -1,-1,-1,  -1,-1, 1,  -1, 1, 1, // v7-v2-v1
    -1,-1,-1,   1,-1,-1,   1,-1, 1, // v7-v4-v3 (bottom)
    1,-1, 1,  -1,-1, 1,  -1,-1,-1,  // v3-v2-v7
    1,-1,-1,  -1,-1,-1,  -1, 1,-1,  // v4-v7-v6 (back)
    -1, 1,-1,   1, 1,-1,   1,-1,-1  // v6-v5-v4
};

// normal array
GLfloat normals[]   = { 0, 0, 1,   0, 0, 1,   0, 0, 1,  // v0-v1-v2 (front)
    0, 0, 1,   0, 0, 1,   0, 0, 1,  // v2-v3-v0
    1, 0, 0,   1, 0, 0,   1, 0, 0,  // v0-v3-v4 (right)
    1, 0, 0,   1, 0, 0,   1, 0, 0,  // v4-v5-v0
    0, 1, 0,   0, 1, 0,   0, 1, 0,  // v0-v5-v6 (top)
    0, 1, 0,   0, 1, 0,   0, 1, 0,  // v6-v1-v0
    -1, 0, 0,  -1, 0, 0,  -1, 0, 0, // v1-v6-v7 (left)
    -1, 0, 0,  -1, 0, 0,  -1, 0, 0, // v7-v2-v1
    0,-1, 0,   0,-1, 0,   0,-1, 0,  // v7-v4-v3 (bottom)
    0,-1, 0,   0,-1, 0,   0,-1, 0,  // v3-v2-v7
    0, 0,-1,   0, 0,-1,   0, 0,-1,  // v4-v7-v6 (back)
    0, 0,-1,   0, 0,-1,   0, 0,-1   // v6-v5-v4
};

// color array
GLfloat colors[]    = { 1, 1, 1,   1, 1, 0,   1, 0, 0,  // v0-v1-v2 (front)
    1, 0, 0,   1, 0, 1,   1, 1, 1,  // v2-v3-v0
    1, 1, 1,   1, 0, 1,   0, 0, 1,  // v0-v3-v4 (right)
    0, 0, 1,   0, 1, 1,   1, 1, 1,  // v4-v5-v0
    1, 1, 1,   0, 1, 1,   0, 1, 0,  // v0-v5-v6 (top)
    0, 1, 0,   1, 1, 0,   1, 1, 1,  // v6-v1-v0
    1, 1, 0,   0, 1, 0,   0, 0, 0,  // v1-v6-v7 (left)
    0, 0, 0,   1, 0, 0,   1, 1, 0,  // v7-v2-v1
    0, 0, 0,   0, 0, 1,   1, 0, 1,  // v7-v4-v3 (bottom)
    1, 0, 1,   1, 0, 0,   0, 0, 0,  // v3-v2-v7
    0, 0, 1,   0, 0, 0,   0, 1, 0,  // v4-v7-v6 (back)
    0, 1, 0,   0, 1, 1,   0, 0, 1   // v6-v5-v4
};


// function pointers for VBO Extension
// Windows needs to get function pointers from ICD OpenGL drivers,
// because opengl32.dll does not support extensions higher than v1.1.
#ifdef _WIN32
PFNGLGENBUFFERSARBPROC            pglGenBuffersARB = 0;             // VBO Name Generation Procedure
PFNGLBINDBUFFERARBPROC            pglBindBufferARB = 0;             // VBO Bind Procedure
PFNGLBUFFERDATAARBPROC            pglBufferDataARB = 0;             // VBO Data Loading Procedure
PFNGLBUFFERSUBDATAARBPROC         pglBufferSubDataARB = 0;          // VBO Sub Data Loading Procedure
PFNGLDELETEBUFFERSARBPROC         pglDeleteBuffersARB = 0;          // VBO Deletion Procedure
PFNGLGETBUFFERPARAMETERIVARBPROC  pglGetBufferParameterivARB = 0;   // return various parameters of VBO
PFNGLMAPBUFFERARBPROC             pglMapBufferARB = 0;              // map VBO procedure
PFNGLUNMAPBUFFERARBPROC           pglUnmapBufferARB = 0;            // unmap VBO procedure

#define glGenBuffersARB           pglGenBuffersARB
#define glBindBufferARB           pglBindBufferARB
#define glBufferDataARB           pglBufferDataARB
#define glBufferSubDataARB        pglBufferSubDataARB
#define glDeleteBuffersARB        pglDeleteBuffersARB
#define glGetBufferParameterivARB pglGetBufferParameterivARB
#define glMapBufferARB            pglMapBufferARB
#define glUnmapBufferARB          pglUnmapBufferARB

#endif

int main(int argc, char **argv) {

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize(screenWidth, screenHeight);
    glutInitWindowPosition(100, 100);
    glutCreateWindow("Indirect Rendering");
    glutDisplayFunc(display);
    glutIdleFunc(idleCB);
    glutReshapeFunc(reshapeCB);
    glutKeyboardFunc(keyboardCB);
    glutMouseFunc(mouseCB);
    glutMotionFunc(mouseMotionCB);

    // OpenGL Options
    glEnable(GL_LIGHTING);
    glEnable(GL_CULL_FACE);
    
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
    
    glClearColor(0, 0, 0, 0);
    glClearDepth(1.0f);
    glDepthFunc(GL_LEQUAL);
    
    // Lights
    GLfloat lightKa[] = {.2f, .2f, .2f, 1.0f};  // ambient light
    GLfloat lightKd[] = {.7f, .7f, .7f, 1.0f};  // diffuse light
    GLfloat lightKs[] = {1, 1, 1, 1};           // specular light
    glLightfv(GL_LIGHT0, GL_AMBIENT, lightKa);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, lightKd);
    glLightfv(GL_LIGHT0, GL_SPECULAR, lightKs);
    
    float lightPos[4] = {0, 0, 20, 1}; // positional light
    glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
    glEnable(GL_LIGHT0);
    
    atexit(exitCB);
    
    // LOAD OBJ
    
    // get OpenGL info
    glInfo glInfo;
    glInfo.getInfo();
    
#ifdef _WIN32
    // check VBO is supported by your video card
    if(glInfo.isExtensionSupported("GL_ARB_vertex_buffer_object"))     {
        // get pointers to GL functions
        glGenBuffersARB = (PFNGLGENBUFFERSARBPROC)wglGetProcAddress("glGenBuffersARB");
        glBindBufferARB = (PFNGLBINDBUFFERARBPROC)wglGetProcAddress("glBindBufferARB");
        glBufferDataARB = (PFNGLBUFFERDATAARBPROC)wglGetProcAddress("glBufferDataARB");
        glBufferSubDataARB = (PFNGLBUFFERSUBDATAARBPROC)wglGetProcAddress("glBufferSubDataARB");
        glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC)wglGetProcAddress("glDeleteBuffersARB");
        glGetBufferParameterivARB = (PFNGLGETBUFFERPARAMETERIVARBPROC)wglGetProcAddress("glGetBufferParameterivARB");
        glMapBufferARB = (PFNGLMAPBUFFERARBPROC)wglGetProcAddress("glMapBufferARB");
        glUnmapBufferARB = (PFNGLUNMAPBUFFERARBPROC)wglGetProcAddress("glUnmapBufferARB");
        
        // check once again VBO extension
        if(glGenBuffersARB && glBindBufferARB && glBufferDataARB && glBufferSubDataARB &&
           glMapBufferARB && glUnmapBufferARB && glDeleteBuffersARB && glGetBufferParameterivARB) {
            vboSupported = vboUsed = true;
            std::cout << "Video card supports GL_ARB_vertex_buffer_object." << std::endl;
        } else {
            vboSupported = vboUsed = false;
            std::cout << "Video card does NOT support GL_ARB_vertex_buffer_object." << std::endl;
        }
    }
    else {
        vboSupported = vboUsed = false;
        std::cout << "Video card does NOT support GL_ARB_vertex_buffer_object." << std::endl;
    }
    
#else
    if(glInfo.isExtensionSupported("GL_ARB_vertex_buffer_object")) {
        vboSupported = vboUsed = true;
        std::cout << "Video card supports GL_ARB_vertex_buffer_object." << std::endl;
    } else {
        vboSupported = vboUsed = false;
        std::cout << "Video card does NOT support GL_ARB_vertex_buffer_object." << std::endl;
    }
#endif
    
//    if(vboSupported) {
//    //  VBO creation code here
//        /* STEPS
//         1. Create the VBO. Note that, you need to delete it when program exits.
//         2. Put the vertex coords, vertex normal and vertex color arrays in the same buffer object.
//         3. Copy data with 3 calls of glBufferSubDataARB, 
//                one for vertex coords
//                one for normals
//                one for colors
//         4. The target flag is GL_ARRAY_BUFFER_ARB, and usage flag is GL_STATIC_DRAW_ARB
//         */
    
    
    
    
//    }
    
    glutMainLoop();
    
    return 0;
}


// Write 2D text on screen
void drawString(const char *str, int x, int y, float color[4], void *font) {
    glDisable(GL_LIGHTING);
    
    glColor4fv(color);
    glRasterPos2i(x, y);
    
    while(*str) {
        glutBitmapCharacter(font, *str);
        ++str;
    }

    glPopAttrib();
}


// Show info on screen
void showInfo() {
    glPushMatrix();                     // save current modelview matrix
    glLoadIdentity();                   // reset modelview matrix
    
    glMatrixMode(GL_PROJECTION);     // switch to projection matrix
    glPushMatrix();                  // save current projection matrix
    glLoadIdentity();                // reset projection matrix
    gluOrtho2D(0, screenWidth, 0, screenHeight);  // set to orthogonal projection
    
    float color[4] = {1, 1, 1, 1};
    
    stringstream ss;
    ss << "VBO: " << (vboUsed ? "on" : "off") << std::ends;  // add 0(ends) at the end
    drawString(ss.str().c_str(), 1, screenHeight-13, color, font);
    ss.str(""); // clear buffer
    
    ss << "Press SPACE to toggle VBO on & off." << std::ends;
    drawString(ss.str().c_str(), 1, 10, color, font);
    
    glPopMatrix();                   // restore to previous projection matrix
    
    glMatrixMode(GL_MODELVIEW);      // switch to modelview matrix
    glPopMatrix();                   // restore to previous modelview matrix
}


void display() {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glPushMatrix();
    
    glTranslatef(0, 0, -cameraDistance);
    glRotatef(cameraAngleX, 1, 0, 0);   // pitch
    glRotatef(cameraAngleY, 0, 1, 0);   // heading
    
//    // EXERCISE 1 - RENDERING CUBE WITH DIRECT RENDERING
    
    
    
    
    
//    // EXERCISE 2 - RENDERING CUBE WITH VERTEX ARRAYS - DIRECT RENDERING
//    /*
//     STEPS
//     2.Enable the Client State
//     2.Specify the pointers to each array
//     3.Draw the elements
//     4.Disable the Client State
//     */

    
    
    
    
    
    
    // EXERCISE 3 - RENDERING CUBE WITH VBO - INDIRECT RENDERING
//    if(vboUsed) {
//          // draw cube using VBO
//    } else {
//         // draw cube with VAO
//    }
    
    
    // EXERCISE 4 - IMPORTING THE EXTERNAL .OBJ FILE
    
    showInfo();
    
    glPopMatrix();
    
    glutSwapBuffers();
}


void reshapeCB(int w, int h) {
    screenWidth = w;
    screenHeight = h;

    glViewport(0, 0, (GLsizei)screenWidth, (GLsizei)screenHeight);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0f, (float)(screenWidth)/screenHeight, 1.0f, 1000.0f); // FOV, AspectRatio, NearClip, FarClip

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}


void idleCB() {
    glutPostRedisplay();
}

void keyboardCB(unsigned char key, int x, int y) {

    switch(key) {
        case 27:
            exit(0);
            break;
            
        case ' ':
            if(vboSupported)
                vboUsed = !vboUsed;
            break;
            
        case 'd':
            drawMode = ++drawMode % 3;
            if(drawMode == 0) {
                glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
                glEnable(GL_DEPTH_TEST);
                glEnable(GL_CULL_FACE);
            } else if(drawMode == 1) {
                glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
                glDisable(GL_DEPTH_TEST);
                glDisable(GL_CULL_FACE);
            } else {
                glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
                glDisable(GL_DEPTH_TEST);
                glDisable(GL_CULL_FACE);
            }
            break;
            
        default:
            ;
    }
}


void mouseCB(int button, int state, int x, int y) {
    mouseX = x;
    mouseY = y;
    
    if(button == GLUT_LEFT_BUTTON) {
        if(state == GLUT_DOWN)
            mouseLeftDown = true;
        else if(state == GLUT_UP)
            mouseLeftDown = false;
    } else if(button == GLUT_RIGHT_BUTTON) {
        if(state == GLUT_DOWN)
            mouseRightDown = true;
        else if(state == GLUT_UP)
            mouseRightDown = false;
    }
}


void mouseMotionCB(int x, int y) {
    if(mouseLeftDown) {
        cameraAngleY += (x - mouseX);
        cameraAngleX += (y - mouseY);
        mouseX = x;
        mouseY = y;
    }
    if(mouseRightDown) {
        cameraDistance -= (y - mouseY) * 0.2f;
        mouseY = y;
    }
}

// Delete the VBO data
void exitCB() {
    if(vboSupported) {
//        glDeleteBuffersARB(1, &vboId);
//        vboId = 0;
    }
}